import numpy as np
cimport numpy as np

from bbhx.utils.utility import pointer_adjust

assert sizeof(int) == sizeof(np.int32_t)

cdef extern from "Interpolate.hh":
    void interpolate(double* freqs, double* propArrays,
                     double* B, double* upper_diag, double* diag, double* lower_diag,
                     int length, int numInterpParams, int numModes, int numBinAll);


@pointer_adjust
def interpolate_wrap(freqs, propArrays,
                     B, upper_diag, diag, lower_diag,
                     length, numInterpParams, numModes, numBinAll):

    cdef size_t freqs_in = freqs
    cdef size_t propArrays_in = propArrays
    cdef size_t B_in = B
    cdef size_t upper_diag_in = upper_diag
    cdef size_t diag_in = diag
    cdef size_t lower_diag_in = lower_diag

    interpolate(<double*>freqs_in, <double*>propArrays_in,
              <double*>B_in, <double*>upper_diag_in, <double*>diag_in, <double*>lower_diag_in,
              length, numInterpParams, numModes, numBinAll)



import numpy as np
cimport numpy as np

from bbhx.utils.utility import pointer_adjust

assert sizeof(int) == sizeof(np.int32_t)

cdef extern from "WaveformBuild.hh":
    ctypedef void* cmplx 'cmplx'

    void InterpTDI(long* templateChannels_ptrs, double* dataFreqs, double* freqs, double* propArrays, double* c1, double* c2, double* c3, double* t_start, double* t_end, int length, int data_length, int numBinAll, int numModes, long* inds_ptrs, int* inds_start, int* ind_lengths);

    void direct_sum(cmplx* templateChannels,
                    double* bbh_buffer,
                    int numBinAll, int data_length, int nChannels, int numModes, double* t_start, double* t_end)


@pointer_adjust
def InterpTDI_wrap(templateChannels_ptrs, dataFreqs, freqs, propArrays, c1, c2, c3, t_start, t_end, length, data_length, numBinAll, numModes, inds_ptrs, inds_start, ind_lengths):

    cdef size_t freqs_in = freqs
    cdef size_t propArrays_in = propArrays
    cdef size_t templateChannels_ptrs_in = templateChannels_ptrs
    cdef size_t dataFreqs_in = dataFreqs
    cdef size_t c1_in = c1
    cdef size_t c2_in = c2
    cdef size_t c3_in = c3
    cdef size_t t_start_in = t_start
    cdef size_t t_end_in = t_end
    cdef size_t inds_ptrs_in = inds_ptrs
    cdef size_t inds_start_in = inds_start
    cdef size_t ind_lengths_in = ind_lengths

    InterpTDI(<long*> templateChannels_ptrs_in, <double*> dataFreqs_in, <double*> freqs_in, <double*> propArrays_in, <double*> c1_in, <double*> c2_in, <double*> c3_in, <double*> t_start_in, <double*> t_end_in, length, data_length, numBinAll, numModes, <long*> inds_ptrs_in, <int*> inds_start_in, <int*> ind_lengths_in);

@pointer_adjust
def direct_sum_wrap(templateChannels,
                bbh_buffer,
                numBinAll, data_length, nChannels, numModes, t_start, t_end):

    cdef size_t templateChannels_in = templateChannels
    cdef size_t bbh_buffer_in = bbh_buffer
    cdef size_t t_start_in = t_start
    cdef size_t t_end_in = t_end

    direct_sum(<cmplx*> templateChannels_in,
                    <double*> bbh_buffer_in,
                    numBinAll, data_length, nChannels, numModes, <double*> t_start_in, <double*> t_end_in)



# Main waveform class location

# Copyright (C) 2021 Michael L. Katz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import warnings

# import gpu stuff
try:
    import cupy as cp
    from .cutils.pyWaveformBuild import direct_sum_wrap as direct_sum_wrap_gpu
    from .cutils.pyWaveformBuild import InterpTDI_wrap as InterpTDI_wrap_gpu

except (ImportError, ModuleNotFoundError) as e:
    print("No CuPy")

from .cutils.pyWaveformBuild_cpu import direct_sum_wrap as direct_sum_wrap_cpu
from .cutils.pyWaveformBuild_cpu import InterpTDI_wrap as InterpTDI_wrap_cpu

from .waveforms.phenomhm import PhenomHMAmpPhase
from .response.fastfdresponse import LISATDIResponse
from .utils.transform import tSSBfromLframe, tLfromSSBframe
from .utils.interpolate import CubicSplineInterpolant
from .utils.constants import *
from .utils.citations import *


class TemplateInterpFD:
    """Interpolate frequency domain template.

    This class wraps :class:`CubicSplineInterpolant <bbhx.utils.interpolate.CubicSplineInterpolant>` so
    that it fits into this specific waveform production method.

    This class has GPU capabilities.

    Args:
        use_gpu (bool, optional): If ``True``, use GPU.

    Attributes:
        data_length (int): Length of data. This class interpolates to this length.
        length (int): Length of original frequency array.
        num_bin_all (int): Number of binaries.
        num_channels (int): Number of channels in data.
        num_modes (int): Number of harmonics.
        template_carrier (complex128 xp.ndarray): Carrier for output templates.
            Templates can be accessed through the ``template_channels`` property.
        use_gpu (bool): If True, using GPU.

    """

    def __init__(self, use_gpu=False):
        self.use_gpu = use_gpu

    @property
    def xp(self) -> object:
        """Numpy or Cupy"""
        return cp if self.use_gpu else np

    @property
    def template_gen(self) -> callable:
        """C/CUDA wrapped function for computing interpolated waveforms"""
        return InterpTDI_wrap_gpu if self.use_gpu else InterpTDI_wrap_cpu

    @property
    def template_channels(self):
        """Get template channels from ``self.template_carrier``."""
        return [
            self.template_carrier[i].reshape(self.num_channels, self.lengths[i])
            for i in range(self.num_bin_all)
        ]

    @property
    def citation(self):
        """citations for this class"""
        return katz_citations

    def __call__(
        self,
        data_freqs,
        interp_container,
        t_start,
        t_end,
        length,
        num_modes,
        num_channels,
    ):
        """Generate frequency domain template via interpolation.

        This class takes all waveform and response information as sparse arrays
        and then interpolates to the proper frequency array.

        This class acts in a unique why by passing arrays of pointers from python
        into C++/CUDA.

        Args:
            data_freqs (double xp.ndarray): Frequencies to interpolate to.
            interp_container (obj): ``container`` attribute from the interpolant
                class: :class:`CubicSplineInterpolant <bbhx.utils.interpolate.CubicSplineInterpolant>`.
            t_start (double xp.ndarray): Array of start times (sec) for each binary.
            t_end (double xp.ndarray): Array of end times (sec) for each binary.
            length (int): Length of original frequency array.
            num_modes (int): Number of harmonics.
            num_channels (int): Number of channels in data.

        Returns:
            list: List of template arrays for all binaries.
                shape of each array: ``(self.num_channels, self.data_length)``

        """

        # fill important quantities
        num_bin_all = len(t_start)
        self.length = length
        self.num_modes = num_modes
        self.num_bin_all = num_bin_all
        self.data_length = len(data_freqs)
        self.num_channels = num_channels

        # unpack interp_container
        (freqs, y, c1, c2, c3) = interp_container

        freqs_shaped = freqs.reshape(self.num_bin_all, self.num_modes, -1)

        # find where each binary's signal starts and ends in the data array

        # lowest frequencies in all the modes. Highest frequencies in all the modes.
        start_and_end = self.xp.asarray(
            [
                freqs_shaped.min(axis=(1, 2)),
                freqs_shaped.max(axis=(1, 2)),
            ]
        ).T

        if self.use_gpu and not isinstance(data_freqs, self.xp.ndarray):
            raise ValueError(
                "Make sure if using Cupy or Numpy, the input freqs array is of the same type."
            )

        inds_start_and_end = self.xp.asarray(
            [
                self.xp.searchsorted(data_freqs, temp, side="right")
                for temp in start_and_end
            ]
        )

        # find proper interpolation window for each point in data stream
        inds = [
            np.concatenate(
                [
                    self.xp.searchsorted(
                        freqs_shaped[i, j], data_freqs[st:et], side="right"
                    )
                    - 1
                    for j in range(self.num_modes)
                ]
            )
            .astype(self.xp.int32)
            .copy()
            for i, (st, et) in enumerate(inds_start_and_end)
        ]

        # lengths of the signals in frequency domain
        self.lengths = lengths = np.asarray(
            [int(len(inds_i) / self.num_modes) for inds_i in inds], dtype=self.xp.int32
        )

        # make sure have this quantity available on CPU
        try:
            temp_inds = inds_start_and_end[:, 0].get()
        except AttributeError:
            temp_inds = inds_start_and_end[:, 0]

        # where to start filling waveform
        self.start_inds = start_inds = (temp_inds.copy()).astype(np.int32)

        # get pointers to all the index arrays of different length
        try:
            self.ptrs = ptrs = np.asarray([ind_i.data.ptr for ind_i in inds])
        except AttributeError:
            self.ptrs = ptrs = np.asarray(
                [ind_i.__array_interface__["data"][0] for ind_i in inds]
            )

        # initialize template information
        self.template_carrier = [
            self.xp.zeros(
                int(self.num_channels * temp_length),
                dtype=self.xp.complex128,
            )
            for temp_length in lengths
        ]

        # get pointers to template carriers so they can be run in streams
        try:
            template_carrier_ptrs = np.asarray(
                [temp_carrier.data.ptr for temp_carrier in self.template_carrier]
            )
        except AttributeError:
            template_carrier_ptrs = np.asarray(
                [
                    temp_carrier.__array_interface__["data"][0]
                    for temp_carrier in self.template_carrier
                ]
            )

        # fill templates
        self.template_gen(
            template_carrier_ptrs,
            data_freqs,
            freqs,
            y,
            c1,
            c2,
            c3,
            t_start,
            t_end,
            self.length,
            self.data_length,
            self.num_bin_all,
            self.num_modes,
            ptrs,
            start_inds,
            lengths,
        )

        # return templates in the right shape
        return self.template_channels


class BBHWaveformFD:
    """Generate waveforms put through response functions

    This class generates waveforms put through the LISA response function. In the
    future, ground-based analysis may be added. Therefore, it currently
    returns the TDI variables according the response keyword arguments given.

    If you use this class, please cite `arXiv:2005.01827 <https://arxiv.org/abs/2005.01827>`_
    and `arXiv:2111.01064 <https://arxiv.org/abs/2111.01064>`_, as well as the papers
    listed for the waveform and response given just below.

    Right now, it is hard coded to produce the waveform with
    :class:`PhenomHMAmpPhase <bbhx.waveforms.phenomhm.PhenomHMAmpPhase>`. This can also be used
    to produce PhenomD. See the docs for that waveform. The papers describing PhenomHM/PhenomD
    waveforms are here: `arXiv:1708.00404 <https://arxiv.org/abs/1708.00404>`_,
    `arXiv:1508.07250 <https://arxiv.org/abs/1508.07250>`_, and
    `arXiv:1508.07253 <https://arxiv.org/abs/1508.07253>`_.

    The response function is the fast frequency domain response function
    from `arXiv:1806.10734 <https://arxiv.org/abs/1806.10734>`_ and
    `arXiv:2003.00357 <https://arxiv.org/abs/2003.00357>`_. It is implemented in
    :class:`LISATDIResponse <bbhx.response.fastfdresponse.LISATDIResponse`.

    This class is GPU accelerated.

    Args:
        amp_phase_kwargs (dict, optional): Keyword arguments for the
            initialization of the ampltidue-phase waveform class: :class:`PhenomHMAmpPhase <bbhx.waveforms.phenomhm.PhenomHMAmpPhase>`.
        response_kwargs (dict, optional): Keyword arguments for the initialization
            of the response class: :class:`LISATDIResponse <bbhx.response.fastfdresponse.LISATDIResponse`.
        interp_kwargs (dict, optional): Keyword arguments for the initialization
            of the interpolation class: :class:`TemplateInterpFD`.
        use_gpu (bool, optional): If ``True``, use a GPU. (Default: ``False``)

    Attributes:
        amp_phase_gen (obj): Waveform generation class.
        data_length (int): Length of the final output data.
        interp_response (obj): Interpolation class.
        length (int): Length of initial evaluations of waveform and response.
        num_bin_all (int): Total number of binaries analyzed.
        num_interp_params (int): Number of parameters to interpolate (9).
        num_modes (int): Number of harmonic modes.
        out_buffer_final (xp.ndarray): Array with buffer information with shape:
            ``(self.num_interp_params, self.num_bin_all, self.num_modes, self.length)``.
            The order of the parameters is amplitude, phase, t-f, transferL1re, transferL1im,
            transferL2re, transferL2im, transferL3re, transferL3im.

    """

    def __init__(
        self,
        amp_phase_kwargs={},
        response_kwargs={},
        interp_kwargs={},
        use_gpu=False,
    ):
        self.use_gpu = use_gpu

        # initialize waveform and response funtions
        self.amp_phase_gen = PhenomHMAmpPhase(**amp_phase_kwargs, use_gpu=use_gpu)
        self.response_gen = LISATDIResponse(**response_kwargs, use_gpu=use_gpu)

        self.num_interp_params = 9

        # setup the final interpolant
        self.interp_response = TemplateInterpFD(**interp_kwargs, use_gpu=use_gpu)

    @property
    def use_gpu(self) -> bool:
        """Whether to use a GPU."""
        return self._use_gpu

    @use_gpu.setter
    def use_gpu(self, use_gpu: bool) -> None:
        """Set ``use_gpu``."""
        assert isinstance(use_gpu, bool)
        self._use_gpu = use_gpu

    @property
    def xp(self) -> object:
        """Numpy or Cupy"""
        return cp if self.use_gpu else np

    @property
    def waveform_gen(self) -> callable:
        """C/CUDA wrapped function for computing waveforms"""
        return direct_sum_wrap_gpu if self.use_gpu else direct_sum_wrap_cpu

    @property
    def citation(self):
        """Citations for this class"""
        return (
            katz_citations + marsat_1 + marsat_2 + phenomhm_citation + phenomd_citations
        )

    def __call__(
        self,
        m1,
        m2,
        chi1z,
        chi2z,
        distance,
        phi_ref,
        f_ref,
        inc,
        lam,
        beta,
        psi,
        t_ref,
        t_obs_start=0.0,
        t_obs_end=1.0,
        freqs=None,
        length=None,
        modes=None,
        shift_t_limits=True,
        direct=False,
        compress=True,
        squeeze=False,
        fill=False,
        combine=False,
    ):
        """Generate the binary black hole frequency-domain TDI waveforms


        Args:
            m1 (double scalar or np.ndarray): Mass 1 in Solar Masses :math:`(m1 > m2)`.
            m2 (double or np.ndarray): Mass 2 in Solar Masses :math:`(m1 > m2)`.
            chi1z (double or np.ndarray): Dimensionless spin 1 (for Mass 1) in Solar Masses.
            chi2z (double or np.ndarray): Dimensionless spin 2 (for Mass 1) in Solar Masses.
            distance (double or np.ndarray): Luminosity distance in m.
            phi_ref (double or np.ndarray): Phase at ``f_ref``.
            f_ref (double or np.ndarray): Reference frequency at which ``phi_ref`` and ``t_ref`` are set.
                If ``f_ref == 0``, it will be set internally by the PhenomHM code
                to :math:`f_\\text{max} = \\text{max}(f^2A_{22}(f))`.
            inc (double or np.ndarray): Inclination of the binary in radians :math:`(\iota\in[0.0, \pi])`.
            lam (double or np.ndarray): Ecliptic longitude :math:`(\lambda\in[0.0, 2\pi])`.
            beta (double or np.ndarray): Ecliptic latitude :math:`(\\beta\in[-\pi/2, \pi/2])`.
            psi (double or np.ndarray): Polarization angle in radians :math:`(\psi\in[0.0, \pi])`.
            t_ref (double or np.ndarray): Reference time in seconds. It is set at ``f_ref``.
            t_obs_start (double, optional): Start time of observation in years
                in the LISA constellation reference frame. This is with reference to :math:`t=0`.
                (Default: 0.0)
            t_obs_end (double, optional): End time of observation in years in the
                LISA constellation reference frame. This is with reference to :math:`t=0`.
                (Default: 1.0)
            freqs (np.ndarray, optional): Frequencies at which to evaluate the final waveform.
                If ``length`` is also given, the interpolants interpolate to these
                frequencies. If ``length`` is not given, the waveform amplitude, phase,
                and response will be directly evaluated at these frequencies. In this case,
                a 2D np.ndarray can also be provided. (Default: ``None``)
            length (int, optional): Number of frequencies to use in sparse array for
                interpolation.
            modes (list, optional): Harmonic modes to use. If not given, they will
                default to those available in the waveform model. For PhenomHM:
                [(2,2), (3,3), (4,4), (2,1), (3,2), (4,3)]. For PhenomD: [(2,2)].
                (Default: ``None``)
            shift_t_limits (bool, optional): If ``False``, ``t_obs_start`` and ``t_obs_end``
                are relative to ``t_ref`` counting backwards in time. If ``True``,
                those quantities are relative to :math:`t=0`. (Default: ``False``)
            direct (bool, optional): If ``True``, directly compute the waveform without
                interpolation. (Default: ``False``)
            compress (bool, optional): If ``True``, combine harmonics into single channel
                waveforms. (Default: ``True``)
            squeeze (bool, optional): If ``True``, remove any axes of length 1 from the
                final return information. (Default: ``False``)
            fill (bool, optional): If ``True``, fill data streams according to the ``combine``
                keyword argument. If ``False, returns information for the fast likelihood functions.
            combine (bool, optional): If ``True``, combine all waveforms into the same output
                data stream. (Default: ``False``)


        Returns:
            xp.ndarray: Shape ``(3, self.length, self.num_bin_all)``.
                Final waveform for each binary. If ``direct==True`` and ``compress==True``.
                # TODO: switch dimensions?
            xp.ndarray:  Shape ``(3, self.num_modes, self.length, self.num_bin_all)``.
                Final waveform for each binary. If ``direct==True`` and ``compress==False``.
            xp.ndarray:  Shape ``(3, self.data_length)``.
                Final waveform of all binaries in the same data stream.
                If ``fill==True`` and ``combine==True``.
            xp.ndarray:  Shape ``(self.num_bin_all, 3, self.data_length)``.
                Final waveform of all binaries in the same data stream.
                If ``fill==True`` and ``combine==False``.
            tuple: Information for fast likelihood functions.
                First entry is ``template_channels`` property from :class:`TemplateInterpFD`.
                Second entry is ``start_inds`` attribute from ``self.interp_response``.
                Third entry is ``lengths`` attribute from ``self.interp_response``.

        Raises:
            ValueError: ``length`` and ``freqs`` not given. Modes are given but not in a list.


        """

        # make sure everything is at least a 1D array
        m1 = np.atleast_1d(m1)
        m2 = np.atleast_1d(m2)
        chi1z = np.atleast_1d(chi1z)
        chi2z = np.atleast_1d(chi2z)
        distance = np.atleast_1d(distance)
        phi_ref = np.atleast_1d(phi_ref)
        inc = np.atleast_1d(inc)
        lam = np.atleast_1d(lam)
        beta = np.atleast_1d(beta)
        psi = np.atleast_1d(psi)
        t_ref = np.atleast_1d(t_ref)

        self.num_bin_all = len(m1)

        # TODO: add sanity checks for t_start, t_end
        # how to set up time limits
        if shift_t_limits is False:
            warnings.warn(
                "Deprecated: shift_t_limits. Previously shift_t_limits defaulted to False. This option is now removed and permanently set to shift_t_limits=True."
            )
            # t_ref_L = tLfromSSBframe(t_ref, lam, beta)

            # # start and end times are defined in the LISA reference frame
            # t_obs_start_L = t_ref_L - t_obs_start * YRSID_SI
            # t_obs_end_L = t_ref_L - t_obs_end * YRSID_SI

            # # convert to SSB frame
            # t_obs_start_SSB = tSSBfromLframe(t_obs_start_L, lam, beta, 0.0)
            # t_obs_end_SSB = tSSBfromLframe(t_obs_end_L, lam, beta, 0.0)

            # # fix zeros and less than zero
            # t_start = (
            #     t_obs_start_SSB if t_obs_start > 0.0 else np.zeros(self.num_bin_all)
            # )
            # t_end = t_obs_end_SSB if t_obs_end > 0.0 else np.zeros_like(t_start)

        # else:
        # start and end times are defined in the LISA reference frame
        t_obs_start_L = t_obs_start * YRSID_SI
        t_obs_end_L = t_obs_end * YRSID_SI

        # convert to SSB frame
        t_obs_start_SSB = tSSBfromLframe(t_obs_start_L, lam, beta, 0.0)
        t_obs_end_SSB = tSSBfromLframe(t_obs_end_L, lam, beta, 0.0)
        t_start = np.atleast_1d(t_obs_start_SSB)
        t_end = np.atleast_1d(t_obs_end_SSB)

        if freqs is None and length is None:
            raise ValueError("Must input freqs or length.")

        # if computing directly, set info
        elif freqs is not None and direct is True:
            if freqs.ndim == 1:
                # set the length of the input
                length = len(freqs)
            else:
                length = freqs.shape[1]

        # this means the frequencies are what needs to be interpolated to
        elif direct is False:
            self.data_length = len(freqs)
            if length is None:
                raise ValueError("If direct is False, length parameter must be given.")

        self.length = length

        # setup harmonic modes
        if modes is None:
            # default mode setup
            self.num_modes = len(self.amp_phase_gen.allowable_modes)
        else:
            if not isinstance(modes, list):
                raise ValueError("modes must be a list.")
            self.num_modes = len(modes)

        self.num_bin_all = len(m1)

        out_buffer = self.xp.zeros(
            (self.num_interp_params * self.length * self.num_modes * self.num_bin_all)
        )

        freqs_temp = freqs if direct else None

        phi_ref_amp_phase = np.zeros_like(m1)

        Tobs = t_end - t_start
        self.amp_phase_gen(
            m1,
            m2,
            chi1z,
            chi2z,
            distance,
            phi_ref_amp_phase,
            f_ref,
            t_ref,
            length,
            freqs=freqs_temp,
            out_buffer=out_buffer,
            modes=modes,
            Tobs=Tobs,
            direct=direct,
        )

        # setup buffer to carry around all the quantities of interest
        # params are amp, phase, tf, transferL1re, transferL1im, transferL2re, transferL2im, transferL3re, transferL3im
        out_buffer = out_buffer.reshape(
            self.num_interp_params, self.num_bin_all, self.num_modes, self.length
        )

        out_buffer = out_buffer.flatten().copy()

        # compute response function
        self.response_gen(
            self.amp_phase_gen.freqs,
            inc,
            lam,
            beta,
            psi,
            phi_ref,
            length,
            out_buffer=out_buffer,  # fill into this buffer
            modes=self.amp_phase_gen.modes,
            direct=direct,
        )

        # for checking
        self.out_buffer_final = out_buffer.reshape(
            9, self.num_bin_all, self.num_modes, self.length
        ).copy()

        # direct computation from buffer
        # + compressing all harmonics into a single data stream by direct combination
        if direct and compress:
            # setup template
            templateChannels = self.xp.zeros(
                (self.num_bin_all * 3 * self.length), dtype=self.xp.complex128
            )

            # direct computation of 3 channel waveform
            self.waveform_gen(
                templateChannels,
                out_buffer,
                self.num_bin_all,
                self.length,
                3,
                self.num_modes,
                self.xp.asarray(t_start),
                self.xp.asarray(t_end),
            )

            out = templateChannels.reshape(self.num_bin_all, 3, self.length)

            if squeeze:
                out = out.squeeze()

            return out

        elif direct:
            out = self.xp.zeros(
                (self.num_bin_all, 3, self.num_modes, self.length),
                dtype=self.xp.complex128,
            )
            for mode_i in range(self.num_modes):
                # setup template
                templateChannels = self.xp.zeros(
                    (self.num_bin_all * 3 * self.length), dtype=self.xp.complex128
                )

                out_buffer_temp = (
                    self.out_buffer_final[:, :, mode_i, :].copy().flatten()
                )
                # direct computation of 3 channel waveform
                self.waveform_gen(
                    templateChannels,
                    out_buffer_temp,
                    self.num_bin_all,
                    self.length,
                    3,
                    1,  # num_modes
                    self.xp.asarray(t_start),
                    self.xp.asarray(t_end),
                )

                out[:, :, mode_i, :] += templateChannels.reshape(
                    self.num_bin_all, 3, self.length
                )

            if squeeze:
                out = out.squeeze()

            return out

        else:

            # setup interpolant
            spline = CubicSplineInterpolant(
                self.amp_phase_gen.freqs,
                out_buffer,
                length=self.length,
                num_interp_params=self.num_interp_params,
                num_modes=self.num_modes,
                num_bin_all=self.num_bin_all,
                use_gpu=self.use_gpu,
            )

            # TODO: try single block reduction for likelihood (will probably be worse for smaller batch, but maybe better for larger batch)?
            template_channels = self.interp_response(
                freqs,
                spline.container,
                t_start,
                t_end,
                self.length,
                self.num_modes,
                3,
            )

            # fill the data stream
            if fill:
                if combine:
                    # combine into one data stream
                    data_out = self.xp.zeros((3, len(freqs)), dtype=self.xp.complex128)
                    for temp, start_i, length_i in zip(
                        template_channels,
                        self.interp_response.start_inds,
                        self.interp_response.lengths,
                    ):
                        data_out[:, start_i : start_i + length_i] += temp

                    if squeeze:
                        return data_out.squeeze()

                    return data_out

                else:
                    # put in separate data streams
                    data_out = self.xp.zeros(
                        (self.num_bin_all, 3, len(freqs)), dtype=self.xp.complex128
                    )
                    for bin_i, (temp, start_i, length_i) in enumerate(
                        zip(
                            template_channels,
                            self.interp_response.start_inds,
                            self.interp_response.lengths,
                        )
                    ):
                        data_out[bin_i, :, start_i : start_i + length_i] = temp

                return data_out
            else:
                # return information for the fast likelihood functions
                return (
                    template_channels,
                    self.interp_response.start_inds,
                    self.interp_response.lengths,
                )



# Interpolant for GPUs

# Copyright (C) 2021 Michael L. Katz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

try:
    import cupy as cp
    from ..cutils.pyInterpolate import interpolate_wrap as interpolate_wrap_gpu


except (ImportError, ModuleNotFoundError) as e:
    print("No CuPy or GPU interpolation available.")

from ..cutils.pyInterpolate_cpu import interpolate_wrap as interpolate_wrap_cpu

from ..utils.constants import *


def searchsorted2d_vec(a, b, xp=None, **kwargs):
    if xp is None:
        xp = np

    m, n = a.shape
    max_num = xp.maximum(a.max() - a.min(), b.max() - b.min()) + 1
    r = max_num * xp.arange(a.shape[0])[:, None]
    p = xp.searchsorted((a + r).ravel(), (b + r).ravel(), **kwargs).reshape(m, -1)

    out = p - n * (xp.arange(m)[:, None])
    try:
        xp.cuda.runtime.deviceSynchronize()
    except AttributeError:
        pass

    return out


class CubicSplineInterpolant:
    """GPU-accelerated Multiple Cubic Splines

    This class produces multiple cubic splines. The cubic splines are produced
    with "not-a-knot" boundary conditions.

    This class has GPU capability.

    Args:
        x (xp.ndarray): f values as input for the spline. Can be 1D flattend array
            of total length
            ``(num_bin_all * length)`` or 2D array with shape: ``(num_bin_all, length)``.
        y_all (xp.ndarray): y values for the spline. This can be a 1D flattened
            array with length
            ``(num_interp_params * num_bin_all * num_modes * length)``
            or 4D arrays of shape: ``(num_interp_params, num_bin_all, num_modes, length)``.
        num_interp_params (int, optional): If ``x`` and ``y_all`` are flattened,
            the user must provide the number of interpolation parameters.
            (Default: ``None``)
        num_bin_all (int, optional): If ``x`` and ``y_all`` are flattened,
            the user must provide the number of total binaries.
            (Default: ``None``)
        num_modes (int, optional): If ``x`` and ``y_all`` are flattened,
            the user must provide the number of modes.
            (Default: ``None``)
        length (int, optional): If ``x`` and ``y_all`` are flattened,
            the user must provide the length of the frequency array for each binary.
            (Default: ``None``)
        use_gpu (bool, optional): If True, prepare arrays for a GPU. Default is
            False.

    Raises:
        ValueError: If input arguments are not correct.

    """

    def __init__(
        self,
        x,
        y_all,
        num_interp_params=None,
        num_bin_all=None,
        num_modes=None,
        length=None,
        use_gpu=False,
    ):

        # check all inputs

        self.use_gpu = use_gpu

        # first check is for flattened arrays
        if x.ndim == 1 or y_all.ndim == 1:
            if x.ndim != 1 or y_all.ndim != 1:
                raise ValueError(
                    "If providing flattened x and y_all, need to both be flattened."
                )
            if (
                length is None
                or num_modes is None
                or num_bin_all is None
                or num_interp_params is None
            ):
                raise ValueError(
                    "If providing flattened arrays, need to provide dimensional information: length, num_modes, num_bin_all, num_interp_params."
                )

            if len(x) != length * num_modes * num_bin_all:
                raise ValueError(
                    f"Length of the x array is not correct. It is supposed to be {length * num_bin_all}. It is currently {len(x)}."
                )
            if len(y_all) != length * num_modes * num_bin_all * num_interp_params:
                raise ValueError(
                    f"Length of the y_all array is not correct. It is supposed to be {length * num_modes * num_bin_all * num_interp_params}. It is currently {len(y_all)}."
                )

        else:
            # arrays are shaped
            if x.ndim != 3:
                raise ValueError(
                    "If providing shaped array, needs to be 3D with shape (num_bin_all, num_modes, length)."
                )
            if y_all.ndim != 4:
                raise ValueError(
                    "If providing shaped array, needs to be 4D with shape (num_interp_params, num_bin_all, num_modes, length)."
                )

            num_interp_params, num_bin_all, num_modes, length = y_all.shape

            assert y_all.shape[1:] == x.shape

            x = x.flatten()
            y_all = y_all.flatten()

        # get/store info
        ninterps = num_modes * num_interp_params * num_bin_all
        self.degree = 3

        self.length = length

        # get reshape information
        self.reshape_shape = (num_interp_params, num_bin_all, num_modes, length)
        self.x_reshape_shape = (num_bin_all, num_modes, length)

        # setup all arrays for interpolation
        x = self.xp.asarray(x)
        B = self.xp.zeros((ninterps * length,))
        self.c1 = upper_diag = self.xp.zeros_like(B)
        self.c2 = diag = self.xp.zeros_like(B)
        self.c3 = lower_diag = self.xp.zeros_like(B)
        self.y = y_all

        # perform interpolation
        self.interpolate_arrays(
            x,
            y_all,
            B,
            upper_diag,
            diag,
            lower_diag,
            length,
            num_interp_params,
            num_modes,
            num_bin_all,
        )

        self.x = x.copy()

    @property
    def use_gpu(self) -> bool:
        """Whether to use a GPU."""
        return self._use_gpu

    @use_gpu.setter
    def use_gpu(self, use_gpu: bool) -> None:
        """Set ``use_gpu``."""
        assert isinstance(use_gpu, bool)
        self._use_gpu = use_gpu

    @property
    def xp(self) -> object:
        """Numpy or Cupy"""
        return cp if self.use_gpu else np

    @property
    def interpolate_arrays(self) -> callable:
        """C/CUDA wrapped function for computing interpolation."""
        return interpolate_wrap_gpu if self.use_gpu else interpolate_wrap_cpu

    @property
    def x_shaped(self):
        """Get shaped x array."""
        return self.x.reshape(self.x_reshape_shape)

    @property
    def y_shaped(self):
        """Get shaped y array."""
        return self.y.reshape(self.reshape_shape)

    @property
    def c1_shaped(self):
        """Get shaped c1 array."""
        return self.c1.reshape(self.reshape_shape)

    @property
    def c2_shaped(self):
        """Get shaped c2 array."""
        return self.c2.reshape(self.reshape_shape)

    @property
    def c3_shaped(self):
        """Get shaped c3 array."""
        return self.c3.reshape(self.reshape_shape)

    @property
    def container(self):
        """Container for easy transit of interpolation information."""
        return [self.x, self.y, self.c1, self.c2, self.c3]

    def __call__(self, x_new):
        if x_new.ndim == 3 and x_new.shape[:2] != self.x_reshape_shape[:2]:
            raise ValueError(
                "x_new must have 3D shape of (num_bin_all, num_modes, # of new points)"
            )

        assert self.xp.all(
            x_new <= self.x_shaped.max(axis=-1)[:, :, None]
        ) and self.xp.all(x_new >= self.x_shaped.min(axis=-1)[:, :, None])

        segment_inds = (
            searchsorted2d_vec(
                self.x_shaped.reshape(-1, self.length),
                x_new.reshape(-1, x_new.shape[-1]),
                xp=self.xp,
                side="right",
            )
            - 1
        ).reshape(x_new.shape)

        if self.xp.any(segment_inds == self.length - 1):
            #  assert self.xp.all(x_new[segment_inds == self.length - 1] == self.x_shaped.max(axis=-1))
            segment_inds[segment_inds == self.length - 1] = self.length - 2

        x0 = self.xp.take_along_axis(self.x_shaped, segment_inds, axis=-1)
        y0 = self.xp.take_along_axis(
            self.y_shaped, segment_inds[None, :, :, :], axis=-1
        )
        c1 = self.xp.take_along_axis(
            self.c1_shaped, segment_inds[None, :, :, :], axis=-1
        )
        c2 = self.xp.take_along_axis(
            self.c2_shaped, segment_inds[None, :, :, :], axis=-1
        )
        c3 = self.xp.take_along_axis(
            self.c3_shaped, segment_inds[None, :, :, :], axis=-1
        )

        dx = x_new - x0

        y_new = y0 + c1 * dx + c2 * dx**2 + c3 * dx**3

        if hasattr(self.xp, "cuda"):
            self.xp.get_default_memory_pool().free_all_block()
        return y_new



# LISA Response Functions

# Copyright (C) 2021 Michael L. Katz, Sylvain Marsat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

try:
    import cupy as cp
    from .cutils.pyFDResponse import LISA_response_wrap as LISA_response_wrap_gpu

except (ImportError, ModuleNotFoundError) as e:
    print("No CuPy or GPU response available.")

from ..cutils.pyFDResponse_cpu import LISA_response_wrap as LISA_response_wrap_cpu
from ..utils.constants import *

from lisatools.detector import EqualArmlengthOrbits, Orbits


class LISATDIResponse:
    """Evaluate the fast frequency domain response function

    The response function is the fast frequency domain response function
    from `arXiv:1806.10734 <https://arxiv.org/abs/1806.10734>`_ and
    `arXiv:2003.00357 <https://arxiv.org/abs/2003.00357>`_. Please cite
    these papers if this class is used. This response assumes a fixed,
    non-breathing armlength for the LISA constellation.

    This class has GPU capability.

    Args:
        TDItag (str, optional): TDI channels to generate. Options are ``"XYZ"`` and
            ``"AET"``. If ``"XYZ"`` is not given, it will default to ``"AET"``.
            (Default: ``"AET"``)
        orbits (Orbits, optional): Orbit class. If ``None``, orbits is set to
            :class:`EqualArmlengthOrbits`. (Default: ``None``)
        rescaled (bool, optional): If ``True``, rescale TDI functions to avoid
            infinities at high frequency. (Default: ``False``)
        tdi2 (bool, optional): If ``True``, apply a factor of :math:`-2i \\sin{(4x)}e^{i4x})`
            to tdi1 output. This is a conversion from TDI 1 to TDI 2 under the assumption of equal armlengt orbits.
            (Default: ``False``)
        order_fresnel_stencil (int, optional): Order of the Fresnel stencil in the
            response. Currently, anything above 0 is not implemented. This is left
            in for future compatibility. (Default: ``0``)
        use_gpu (bool, optional): If ``True``, use a GPU. (Default: ``False``)

    Attributes:
        allowable_modes (list): Allowed list of mode tuple pairs ``(l,m)`` for
            the chosen waveform model.
        ells_default (np.ndarray): Default values for the ``l`` index of the harmonic.
        mms_default (np.ndarray): Default values for the ``m`` index of the harmonic.
        includes_amps (bool): If ``True``, the ``out_buffer`` contains the first
            entry for amplitudes.
        order_fresnel_stencil (int): Order of the Fresnel stencil in the
            response. Currently, anything above 0 is not implemented. This is left
            in for future compatibility.
        TDItag (str): TDI channels to generate. Either ``"XYZ"`` or ``"AET"``.

    """

    def __init__(
        self,
        TDItag="AET",
        orbits: Orbits | None = None,
        rescaled: bool = False,
        tdi2: bool = False,
        order_fresnel_stencil=0,
        use_gpu=False,
    ):
        self.rescaled = rescaled
        self.tdi2 = tdi2

        # gpu setup
        self.use_gpu = use_gpu

        if order_fresnel_stencil > 0:
            raise NotImplementedError

        self.order_fresnel_stencil = order_fresnel_stencil

        # TDI setup
        self.TDItag = TDItag
        if TDItag == "XYZ":
            self.TDItag_int = 1

        else:
            self.TDItag_int = 2

        # PhenomHM modes
        self.allowable_modes = [(2, 2), (3, 3), (4, 4), (2, 1), (3, 2), (4, 3)]

        self.ells_default = self.xp.array([2, 3, 4, 2, 3, 4], dtype=self.xp.int32)

        self.mms_default = self.xp.array([2, 3, 4, 1, 2, 3], dtype=self.xp.int32)

        self.orbits = orbits

    @property
    def use_gpu(self) -> bool:
        """Whether to use a GPU."""
        return self._use_gpu

    @use_gpu.setter
    def use_gpu(self, use_gpu: bool) -> None:
        """Set ``use_gpu``."""
        assert isinstance(use_gpu, bool)
        self._use_gpu = use_gpu

    @property
    def response_gen(self):
        """C function on GPU/CPU"""
        response_gen = (
            LISA_response_wrap_gpu if self.use_gpu else LISA_response_wrap_cpu
        )
        return response_gen

    @property
    def xp(self):
        """Numpy or Cupy"""
        xp = cp if self.use_gpu else np
        return xp

    @property
    def orbits(self) -> Orbits:
        return self._orbits

    @orbits.setter
    def orbits(self, orbits: Orbits) -> None:
        if orbits is None:
            self._orbits = EqualArmlengthOrbits()
        elif not isinstance(orbits, Orbits):
            raise ValueError(
                "Input orbits must be of type Orbits (from LISA Analysis Tools)."
            )
        else:
            self._orbits = orbits
        self._orbits.configure(linear_interp_setup=True)

    @property
    def citation(self):
        """Return citations for this class"""
        return katz_citations + marsat_1 + marsat_2

    def _sanity_check_modes(self, ells, mms):
        """Make sure modes are allowed"""
        for ell, mm in zip(ells, mms):
            if (ell, mm) not in self.allowable_modes:
                raise ValueError(
                    "Requested mode [(l,m) = ({},{})] is not available. Allowable modes include {}".format(
                        ell, mm, self.allowable_modes
                    )
                )

    def _initialize_response_container(self):
        """setup reponse container if needed"""
        self.response_carrier = self.xp.zeros(
            (self.nparams * self.length * self.num_modes * self.num_bin_all),
            dtype=self.xp.float64,
        )

    @property
    def transferL1(self):
        """TransferL1 term in response. Shape: ``(num_bin_all, num_modes, length)``"""
        temp = self.response_carrier[
            (self.includes_amps + 2)
            * self.num_per_param : (self.includes_amps + 3)
            * self.num_per_param
        ].reshape(
            self.num_bin_all, self.num_modes, self.length
        ) + 1j * self.response_carrier[
            (self.includes_amps + 3)
            * self.num_per_param : (self.includes_amps + 4)
            * self.num_per_param
        ].reshape(
            self.num_bin_all, self.num_modes, self.length
        )

        return temp

    @property
    def transferL2(self):
        """TransferL2 term in response. Shape: ``(num_bin_all, num_modes, length)``"""
        temp = self.response_carrier[
            (self.includes_amps + 4)
            * self.num_per_param : (self.includes_amps + 5)
            * self.num_per_param
        ].reshape(
            self.num_bin_all, self.num_modes, self.length
        ) + 1j * self.response_carrier[
            (self.includes_amps + 5)
            * self.num_per_param : (self.includes_amps + 6)
            * self.num_per_param
        ].reshape(
            self.num_bin_all, self.num_modes, self.length
        )

        return temp

    @property
    def transferL3(self):
        """TransferL3 term in response. Shape: ``(num_bin_all, num_modes, length)``"""
        temp = self.response_carrier[
            (self.includes_amps + 6)
            * self.num_per_param : (self.includes_amps + 7)
            * self.num_per_param
        ].reshape(
            self.num_bin_all, self.num_modes, self.length
        ) + 1j * self.response_carrier[
            (self.includes_amps + 7)
            * self.num_per_param : (self.includes_amps + 8)
            * self.num_per_param
        ].reshape(
            self.num_bin_all, self.num_modes, self.length
        )

        return temp

    @property
    def phase(self):
        """Get updated phase info. Shape: ``(num_bin_all, num_modes, length)``"""
        phase = self.response_carrier[
            (self.includes_amps + 0)
            * self.num_per_param : (self.includes_amps + 1)
            * self.num_per_param
        ].reshape(self.num_bin_all, self.num_modes, self.length)

        return phase

    @property
    def tf(self):
        """Get tf info. Shape: ``(num_bin_all, num_modes, length)``"""
        tf = self.response_carrier[
            (self.includes_amps + 1)
            * self.num_per_param : (self.includes_amps + 2)
            * self.num_per_param
        ].reshape(self.num_bin_all, self.num_modes, self.length)

        return tf

    def __call__(
        self,
        freqs,
        inc,
        lam,
        beta,
        psi,
        phi_ref,
        length,
        modes=None,
        phase=None,
        tf=None,
        out_buffer=None,
        adjust_phase=True,
        direct=False,
    ):
        """Evaluate respones function

        Args:
            freqs (1D or 2D xp.ndarray): Frequency at which the response is evaluated.
                2D shape is ``(num_bin_all, length)``. If given as a 1D array,
                it should be of length ``num_bin_all * length``.
            inc (scalar or 1D xp.ndarray): Inclination of BBH system in radians.
            lam (scalar or 1D xp.ndarray): Ecliptic longitude in SSB frame in radians.
            beta (scalar or 1D xp.ndarray): Ecliptic latitude in SSB frame in radians.
            psi (scalar or 1D xp.ndarray): Polarization angle of the system in radians.
            phi_ref (scalar or 1D xp.ndarray): Reference phase. **Note**:
                The response function rotates the source by ``phi_ref``. For this reason,
                the main waveform functions (e.g. :class:`bbhx.waveform.BBHWaveformFD`)
                provide ``phi_ref = 0.0`` into the source-frame scaled waveform generators
                (e.g. :class:`bbhx.waveforms.phenomhm.PhenomHMAmpPhase`). This allows
                the reference phase to be applied here in the response.
            length (int): The length of the individual frequency arrays. This is required
                because of the options for putting in 1D arrays into this function.
                The length tells the chunk size in a 1D array.
            modes (list, optional): Harmonic modes to use. If not given, they will
                default to those available in the waveform model PhenomHM:
                ``[(2,2), (3,3), (4,4), (2,1), (3,2), (4,3)]``. (Default: ``None``)
            phase (xp.ndarray, optional): Waveform phase. This is adjusted by the ``phaseRdelay``
                quantity in the code. If more than 1D, the shape should be
                ``(num_bin_all, num_modes, length)``. If 1D, its total length
                should be equivalent to ``num_bin_all * num_modes * length``.
                If ``out_buffer`` is not provided, ``phase`` and ``tf`` are required.
            tf (xp.ndarray, optional): Waveform time-frequency correspondence. This tells the
                response where the LISA constellation is at each frequency.
                If more than 1D, the shape should be
                ``(num_bin_all, num_modes, length)``. If 1D, its total length
                should be equivalent to ``num_bin_all * num_modes * length``.
                If ``out_buffer`` is not provided, ``phase`` and ``tf`` are required.
            out_buffer (xp.ndarray, optional): 1D array initialized to contain all computations
                from the inital waveform and response function. If providing ``out_buffer``,
                the response fills it directly. To make this happen easily in GPU/CPU
                agnostic manner, out_buffer needs to be a 1D array with length
                equivalent to ``nparams * num_bin_all * num_modes * length``.
                ``nparams`` can be 8 if the buffer does not include the amplitudes
                (which are not needed at all for the response computation) or 9
                if it includes the amplitudes. (Default: ``None``)
            adjust_phase (bool, optional): If ``True`` adjust the phase array in-place
                inside the response code. **Note**: This only applies when
                inputing ``phase`` and ``tf``. (Default: ``True``)

        Raises:
            ValueError: Incorrect dimensions for the arrays.

        """

        # to cupy if needed
        inc = self.xp.asarray(self.xp.atleast_1d(inc)).copy()
        lam = self.xp.asarray(self.xp.atleast_1d(lam)).copy()
        beta = self.xp.asarray(self.xp.atleast_1d(beta)).copy()
        psi = self.xp.asarray(self.xp.atleast_1d(psi)).copy()
        phi_ref = self.xp.asarray(self.xp.atleast_1d(phi_ref)).copy()

        # mode setup
        if modes is not None:
            ells = self.xp.asarray([ell for ell, mm in modes], dtype=self.xp.int32)
            mms = self.xp.asarray([mm for ell, mm in modes], dtype=self.xp.int32)
            self._sanity_check_modes(ells, mms)

        else:
            ells = self.ells_default
            mms = self.mms_default

        self.modes = [(ell, mm) for ell, mm in zip(ells, mms)]

        num_modes = len(ells)
        num_bin_all = len(inc)

        # store all info
        self.length = length
        self.num_modes = num_modes
        self.num_bin_all = num_bin_all
        self.num_per_param = length * num_modes * num_bin_all
        self.num_per_bin = length * num_modes

        # number of respones-specific parameters
        self.nresponse_params = 6

        if out_buffer is None and phase is None and tf is None:
            raise ValueError("Must provide either out_buffer or both phase and tf.")

        # setup out_buffer based on inputs
        if out_buffer is None:
            includes_amps = 0
            self.nparams = includes_amps + 2 + self.nresponse_params
            self._initialize_response_container()

        else:
            # use other shape information already known to
            # make sure the given out_buffer is the right length
            # and has the right number of parameters (8 or 9 (including amps))
            nparams_empirical = len(out_buffer) / (
                self.num_bin_all * self.num_modes * self.length
            )

            # indicate if there is an integer number of params in the out_buffer
            if np.allclose(nparams_empirical, np.round(nparams_empirical)) and int(
                nparams_empirical
            ) in [8, 9]:
                self.nparams = int(nparams_empirical)
            else:
                raise ValueError(
                    f"out_buffer incorrect length. The length should be equivalent to (8 or 9) * {self.num_bin_all * self.num_modes * self.length}. Given length is {len(out_buffer)}."
                )

            includes_amps = 1 if self.nparams == 9 else 0
            self.response_carrier = out_buffer

        # if amps are included they are in teh first slot in the array
        self.includes_amps = includes_amps

        # setup and check frequency dimensions
        self.freqs = freqs
        if self.freqs.ndim > 1:
            if self.freqs.shape != (self.num_bin_all, self.num_modes, self.length):
                raise ValueError(
                    f"freqs have incorrect shape. Shape should be {(self.num_bin_all, self.num_modes, self.length)}. Current shape is {freqs.shape}."
                )
            self.freqs = self.freqs.flatten()
        else:
            if len(freqs) != self.num_bin_all * self.num_modes * self.length:
                raise ValueError(
                    f"freqs incorrect length. The length should be equivalent to {self.num_bin_all * self.num_modes * self.length}. Given length is {len(freqs)}."
                )

        # if using phase/tf
        if phase is not None and tf is not None:
            use_phase_tf = True

            if not direct and (
                tf.min() < self.orbits.t_base.min()
                or tf.max() > self.orbits.t_base.max()
            ):
                raise ValueError(
                    f"Orbital information does not cover minimum ({tf.min()}) and maximum ({tf.max()}) tf. Orbital information begins at {self.orbits.t_base.min()} and ends at {self.orbits.t_base.max()}."
                )

            if phase.shape != tf.shape:
                raise ValueError(
                    "Shape of phase array and tf array need to be the same shape."
                )

            if phase.ndim > 1:
                if phase.shape != (self.num_bin_all, self.num_modes, self.length):
                    raise ValueError(
                        f"phase have incorrect shape. Shape should be {(self.num_bin_all, self.num_modes, self.length)}. Current shape is {phase.shape}."
                    )
                # will need to write the phase to original array later if adjust_phase == True
                first = phase.copy().flatten()
                second = tf.copy().flatten()

            else:
                if len(phase) != self.num_bin_all * self.num_modes * self.length:
                    raise ValueError(
                        f"phase incorrect length. The length should be equivalent to {self.num_bin_all * self.num_modes * self.length}. Given length is {len(phase)}."
                    )

                # will need to write the phase to original array later if adjust_phase == True
                first = phase
                second = tf

            # fill the phase into the buffer (which is flat)
            self.response_carrier[
                (includes_amps + 0)
                * length
                * num_modes
                * num_bin_all : (includes_amps + 1)
                * length
                * num_modes
                * num_bin_all
            ] = first

            # fill tf in the buffer (which is flat)
            self.response_carrier[
                (includes_amps + 1)
                * length
                * num_modes
                * num_bin_all : (includes_amps + 2)
                * length
                * num_modes
                * num_bin_all
            ] = second

        elif phase is not None or tf is not None:
            raise ValueError("If provided phase or tf, need to provide both.")

        else:
            use_phase_tf = False
            if not direct and (
                self.tf.min() < self.orbits.t_base.min()
                or self.tf.max() > self.orbits.t_base.max()
            ):
                breakpoint()
                raise ValueError(
                    f"Orbital information does not cover minimum ({self.tf.min()}) and maximum ({self.tf.max()}) tf. Orbital information begins at {self.orbits.t_base.min()} and ends at {self.orbits.t_base.max()}."
                )

        # run response code in C/CUDA
        self.response_gen(
            self.response_carrier,
            ells,
            mms,
            self.freqs,
            phi_ref,
            inc,
            lam,
            beta,
            psi,
            self.TDItag_int,
            self.rescaled,
            self.tdi2,
            self.order_fresnel_stencil,
            num_modes,
            length,
            num_bin_all,
            includes_amps,
            self.orbits,
        )

        # adjust input phase arrays in-place
        if use_phase_tf and adjust_phase:
            output = self.response_carrier[
                (includes_amps + 0)
                * length
                * num_modes
                * num_bin_all : (includes_amps + 1)
                * length
                * num_modes
                * num_bin_all
            ]

            if phase.ndim > 1:
                phase[:] = output.reshape(phase.shape)
            else:
                phase[:] = output



# from future.utils import iteritems
import os
import sys
from os.path import join as pjoin
from setuptools import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy
import argparse


def find_in_path(name, path):
    """Find a file in a search path"""

    # Adapted fom http://code.activestate.com/recipes/52224
    for dir in path.split(os.pathsep):
        binpath = pjoin(dir, name)
        if os.path.exists(binpath):
            return os.path.abspath(binpath)
    return None


def locate_cuda():
    """Locate the CUDA environment on the system

    Returns a dict with keys 'home', 'nvcc', 'include', and 'lib64'
    and values giving the absolute path to each directory.

    Starts by looking for the CUDAHOME env variable. If not found,
    everything is based on finding 'nvcc' in the PATH.
    """

    # First check if the CUDAHOME env variable is in use
    if "CUDAHOME" in os.environ:
        home = os.environ["CUDAHOME"]
        nvcc = pjoin(home, "bin", "nvcc")

    elif "CUDA_HOME" in os.environ:
        home = os.environ["CUDA_HOME"]
        nvcc = pjoin(home, "bin", "nvcc")

    else:
        # Otherwise, search the PATH for NVCC
        nvcc = find_in_path("nvcc", os.environ["PATH"])
        if nvcc is None:
            raise EnvironmentError(
                "The nvcc binary could not be "
                "located in your $PATH. Either add it to your path, "
                "or set $CUDAHOME"
            )
        home = os.path.dirname(os.path.dirname(nvcc))

    cudaconfig = {
        "home": home,
        "nvcc": nvcc,
        "include": pjoin(home, "include"),
        "lib64": pjoin(home, "lib64"),
    }
    for k, v in iter(cudaconfig.items()):
        if not os.path.exists(v):
            raise EnvironmentError(
                "The CUDA %s path could not be " "located in %s" % (k, v)
            )

    return cudaconfig


def customize_compiler_for_nvcc(self):

    # track all the object files generated with cuda device code
    self.cuda_object_files = []

    # Tell the compiler it can processes .cu
    self.src_extensions.append(".cu")

    # Save references to the default compiler_so and _comple methods
    default_compiler_so = self.compiler_so
    super = self._compile

    def _compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
        # generate a special object file that will contain linked in
        # relocatable device code
        if src == "zzzzzzzzzzzzzzzz.cu":
            self.set_executable("compiler_so", CUDA["nvcc"])
            postargs = extra_postargs["nvcclink"]
            cc_args = self.cuda_object_files[1:]
            src = self.cuda_object_files[0]
        elif os.path.splitext(src)[1] == ".cu":
            self.set_executable("compiler_so", CUDA["nvcc"])
            postargs = extra_postargs["nvcc"]
            self.cuda_object_files.append(obj)
        else:
            postargs = extra_postargs["gcc"]
        super(obj, src, ext, cc_args, postargs, pp_opts)
        self.compiler_so = default_compiler_so

    self._compile = _compile


# Run the customize_compiler
class custom_build_ext(build_ext):
    def build_extensions(self):
        customize_compiler_for_nvcc(self.compiler)
        build_ext.build_extensions(self)


try:
    CUDA = locate_cuda()
    run_cuda_install = True
except OSError:
    run_cuda_install = False

parser = argparse.ArgumentParser()

parser.add_argument(
    "--lapack_lib",
    help="Directory of the lapack lib.",
    default="/usr/local/opt/lapack/lib",
)

parser.add_argument(
    "--lapack_include",
    help="Directory of the lapack include.",
    default="/usr/local/opt/lapack/include",
)

parser.add_argument(
    "--lapack",
    help="Directory of both lapack lib and include. '/include' and '/lib' will be added to the end of this string.",
)

parser.add_argument(
    "--gsl_lib", help="Directory of the gsl lib.", default="/usr/local/opt/gsl/lib"
)

parser.add_argument(
    "--gsl_include",
    help="Directory of the gsl include.",
    default="/usr/local/opt/gsl/include",
)

parser.add_argument(
    "--gsl",
    help="Directory of both gsl lib and include. '/include' and '/lib' will be added to the end of this string.",
)

args, unknown = parser.parse_known_args()

for key in [
    args.gsl_include,
    args.gsl_lib,
    args.gsl,
    "--gsl",
    "--gsl_include",
    "--gsl_lib",
    args.lapack_include,
    args.lapack_lib,
    args.lapack,
    "--lapack",
    "--lapack_lib",
    "--lapack_include",
]:
    try:
        sys.argv.remove(key)
    except ValueError:
        pass

# Obtain the numpy include directory. This logic works across numpy versions.
try:
    numpy_include = numpy.get_include()
except AttributeError:
    numpy_include = numpy.get_numpy_include()

if args.lapack is None:
    lapack_include = [args.lapack_include]
    lapack_lib = [args.lapack_lib]

else:
    lapack_include = [args.lapack + "/include"]
    lapack_lib = [args.lapack + "/lib"]

if args.gsl is None:
    gsl_include = [args.gsl_include]
    gsl_lib = [args.gsl_lib]

else:
    gsl_include = [args.gsl + "/include"]
    gsl_lib = [args.gsl + "/lib"]

import lisatools

path_to_lisatools = lisatools.__file__.split("__init__.py")[0]
path_to_lisatools_cutils = path_to_lisatools + "cutils/"

# try:
#     exec(open("scripts/prebuild.py", "r").read())
# except FileNotFoundError:
#     import warnings

#     warnings.warn(
#         "Trying to executre prebuild.py inside setup script, but getting FileNotFoundError. Assuming user will run scripts/prebuild.py manually."
#     )
# if installing for CUDA, build Cython extensions for gpu modules
if run_cuda_install:
    gpu_extension = dict(
        libraries=["cudart", "cublas", "cusparse", "gsl", "gslcblas"],
        library_dirs=[CUDA["lib64"]] + gsl_lib,
        runtime_library_dirs=[CUDA["lib64"]],
        language="c++",
        # This syntax is specific to this build system
        # we're only going to use certain compiler args with nvcc
        # and not with gcc the implementation of this trick is in
        # customize_compiler()
        extra_compile_args={
            "gcc": ["-std=c++11"],  # '-g'],
            "nvcc": [
                "-arch=sm_80",
                # "-gencode=arch=compute_50,code=sm_50",
                # "-gencode=arch=compute_52,code=sm_52",
                "-gencode=arch=compute_60,code=sm_60",
                "-gencode=arch=compute_61,code=sm_61",
                "-gencode=arch=compute_70,code=sm_70",
                "-gencode=arch=compute_75,code=sm_75",
                "-gencode=arch=compute_80,code=compute_80",
                "-std=c++11",
                "-c",
                "--compiler-options",
                "'-fPIC'",
                # "-G",
                # "-g",
                # "-O0",
                # "-lineinfo",
            ],  # for debugging
        },
        include_dirs=[
            numpy_include,
            CUDA["include"],
            "bbhx/cutils/include",
            "/usr/include",
        ],
    )

    pyPhenomHM_ext = Extension(
        "bbhx.cutils.pyPhenomHM",
        sources=["bbhx/cutils/src/PhenomHM.cu", "bbhx/cutils/src/phenomhm.pyx"],
        **gpu_extension,
    )
    pyFDResponse_ext = Extension(
        "bbhx.cutils.pyFDResponse",
        sources=[
            path_to_lisatools_cutils + "src/Detector.cu",
            "bbhx/cutils/src/Response.cu",
            "bbhx/cutils/src/response.pyx",
            "zzzzzzzzzzzzzzzz.cu",
        ],
        libraries=["cudart", "cudadevrt", "cublas", "cusparse"],
        library_dirs=[CUDA["lib64"]],
        runtime_library_dirs=[CUDA["lib64"]],
        language="c++",
        # This syntax is specific to this build system
        # we're only going to use certain compiler args with nvcc
        # and not with gcc the implementation of this trick is in
        # customize_compiler()
        extra_compile_args={
            "gcc": ["-std=c++11"],
            "nvcc": ["-arch=sm_80", "-rdc=true", "--compiler-options", "'-fPIC'"],
            "nvcclink": [
                "-arch=sm_80",
                "--device-link",
                "--compiler-options",
                "'-fPIC'",
            ],
        },
        include_dirs=[
            numpy_include,
            CUDA["include"],
            "bbhx/cutils/include",
            path_to_lisatools_cutils + "include",
            "/usr/include",
        ],
    )
    pyInterpolate_ext = Extension(
        "bbhx.cutils.pyInterpolate",
        sources=["bbhx/cutils/src/Interpolate.cu", "bbhx/cutils/src/interpolate.pyx"],
        **gpu_extension,
    )
    pyWaveformBuild_ext = Extension(
        "bbhx.cutils.pyWaveformBuild",
        sources=[
            "bbhx/cutils/src/WaveformBuild.cu",
            "bbhx/cutils/src/waveformbuild.pyx",
        ],
        **gpu_extension,
    )
    pyLikelihood_ext = Extension(
        "bbhx.cutils.pyLikelihood",
        sources=["bbhx/cutils/src/Likelihood.cu", "bbhx/cutils/src/likelihood.pyx"],
        **gpu_extension,
    )

    # gpu_extensions.append(Extension(extension_name, **temp_dict))

cpu_extension = dict(
    libraries=["lapacke", "lapack", "gsl", "gslcblas"],
    language="c++",
    # This syntax is specific to this build system
    # we're only going to use certain compiler args with nvcc
    # and not with gcc the implementation of this trick is in
    # customize_compiler()
    extra_compile_args={
        "gcc": ["-std=c++11"],
    },  # '-g'],
    include_dirs=[
        numpy_include,
        "bbhx/cutils/include",
        path_to_lisatools_cutils + "include",
        "/usr/include",
    ],
)

pyPhenomHM_cpu_ext = Extension(
    "bbhx.cutils.pyPhenomHM_cpu",
    sources=["bbhx/cutils/src/PhenomHM.cpp", "bbhx/cutils/src/phenomhm_cpu.pyx"],
    **cpu_extension,
)
pyFDResponse_cpu_ext = Extension(
    "bbhx.cutils.pyFDResponse_cpu",
    sources=[
        path_to_lisatools_cutils + "src/Detector.cpp",
        "bbhx/cutils/src/Response.cpp",
        "bbhx/cutils/src/response_cpu.pyx",
    ],
    **cpu_extension,
)

pyInterpolate_cpu_ext = Extension(
    "bbhx.cutils.pyInterpolate_cpu",
    sources=["bbhx/cutils/src/Interpolate.cpp", "bbhx/cutils/src/interpolate_cpu.pyx"],
    **cpu_extension,
)

pyWaveformBuild_cpu_ext = Extension(
    "bbhx.cutils.pyWaveformBuild_cpu",
    sources=[
        "bbhx/cutils/src/WaveformBuild.cpp",
        "bbhx/cutils/src/waveformbuild_cpu.pyx",
    ],
    **cpu_extension,
)

pyLikelihood_cpu_ext = Extension(
    "bbhx.cutils.pyLikelihood_cpu",
    sources=["bbhx/cutils/src/Likelihood.cpp", "bbhx/cutils/src/likelihood_cpu.pyx"],
    **cpu_extension,
)


extensions = [
    pyPhenomHM_cpu_ext,
    pyFDResponse_cpu_ext,
    pyInterpolate_cpu_ext,
    pyWaveformBuild_cpu_ext,
    pyLikelihood_cpu_ext,
]

if run_cuda_install:
    extensions = [
        pyPhenomHM_ext,
        pyFDResponse_ext,
        pyInterpolate_ext,
        pyWaveformBuild_ext,
        pyLikelihood_ext,
    ] + extensions

setup(
    name="bbhx",
    author="Michael Katz",
    author_email="mikekatz04@gmail.com",
    ext_modules=extensions,
    packages=[
        "bbhx",
        "bbhx.utils",
        "bbhx.waveforms",
        "bbhx.response",
        "bbhx.cutils",
        "bbhx.cutils.src",
        "bbhx.cutils.include",
    ],
    # Inject our custom trigger
    cmdclass={"build_ext": custom_build_ext},
    # Since the package has c code, the egg cannot be zipped
    zip_safe=False,
    version="1.1.11",
    python_requires=">=3.6",
    package_data={
        "bbhx.cutils.src": [
            "Interpolate.cu",
            "Interpolate.cpp",
            "Likelihood.cu",
            "Likelihood.cpp",
            "PhenomHM.cu",
            "PhenomHM.cpp",
            "pycppdetector.pyx",
            "Response.cu",
            "Response.cpp",
            "WaveformBuild.cu",
            "WaveformBuild.cpp",
            "interpolate.pyx",
            "likelihood.pyx",
            "phenomhm.pyx",
            "response.pyx",
            "waveformbuild.pyx",
        ],
        "bbhx.cutils.include": [
            "Interpolate.hh",
            "Likelihood.hh",
            "PhenomHM.hh",
            "Response.hh",
            "WaveformBuild.hh",
            "constants.h",
            "cuda_complex.hpp",
            "global.h",
        ],
    },
)
